# =================================================== #
#                                                     #
# @file: RoadNetworkModel.py                          #
# @author: Chris Vesters                              #
# @date: 13/1/2013                                    #
#                                                     #
# Note: it may be helpful to set the path to pydevs.  #
# This can be done by:                                #
# import sys                                          #
# import os.path                                      #
# sys.path.append(os.path...)                         #
#                                                     #
# =================================================== #


# Import code for DEVS model representation:
import pydevs
from pydevs.devs_exceptions import *
from pydevs.infinity import *
from pydevs.DEVS import *

# Import for uniform random number generators
from random import uniform
from random import randint

# =================================================== #
#                                                     #
# Used metrics:                                       #
#   Speed: m/s                                        #
#   Distance: m                                       #
#                                                     #
# Quick reference:                                    #
#   30km/h = 8.3m/s                                   #
#   50km/h = 13.8m/s                                  #
#                                                     #
# =================================================== #

"""
"
" Encapsulates the car.
" This is a modified version of the car used in a project of the course MoSIS.
"
"""
class Car:

  uniqueID = 1;

  """
  "
  " Initialize a car.
  "
  " @param dv_pos_max: The maximum positive acceleration.
  " @param dv_neg_max: The maximum negative acceleration.
  " @param trajectory: The trajectory (list of intersections IDs) the car will follow.  
  "
  """
  def __init__(self, trajectory, dv_pos_max = 10, dv_neg_max = 10):
    self.id = Car.uniqueID
    Car.uniqueID += 1
    self.dv_pos_max = dv_pos_max
    self.dv_neg_max = dv_neg_max
    self.v = 0;
    self.trajectory = trajectory
    

# Define MACROs used by Road:
NONE           = 0
BREAKING_POINT = 1
CHECKING_POINT = 2
SPEED_CHANGE   = 3
END_OF_ROAD    = 4

"""
"
" This class encapsulates the Road.
"
"""
class Road(AtomicDEVS):
  
  """
  "
  " Initiator
  "
  " @param intA: A first intersection.
  " @param intB: A second intersection.
  " @param oneWay: Indicating whether it is a one-way road or not.
  " @param length: The length of the road.
  " @param maxSpeed: The maximum allowed speed.
  "
  """
  def __init__(self, intA, intB, length, maxSpeed, name = "None"):
    AtomicDEVS.__init__(self, name = name)
    
    self.intA = intA
    self.intB = intB
    
    self.car_in = self.addInPort(name="Car_in")
    self.car_out = self.addOutPort(name="Car_out")
    self.query = self.addOutPort(name="Query")
    self.reply = self.addInPort(name="Reply")
    
    self.cars = []
    self.lane = {}
    
    self.length = length
    self.maxSpeed = maxSpeed
    
    self.time_updates = [[INFINITY, 0, NONE]]
  
  
  """
  "
  " The external transition function
  "
  """
  def extTransition(self):
    car = self.peek(self.car_in)
    reply = self.peek(self.reply)
    
    for update in self.time_updates:
      update[0] -= self.elapsed
    
    if car != None:
      self.cars.append(car)
      self.lane[car.id] = self.length
      # Determine the time it wil take for the car to accelerate.
      v_diff = self.maxSpeed - car.v
      if v_diff < 0:
        t = v_diff / car.dv_neg_max
        self.time_updates.append([0, car.id, SPEED_CHANGE, (car.v - self.maxSpeed)/2])
        self.time_updates.append([t, car.id, SPEED_CHANGE, self.maxSpeed])
      elif v_diff > 0:
        t = v_diff / car.dv_pos_max
        self.time_updates.append([0, car.id, SPEED_CHANGE, (self.maxSpeed - car.v)/2])
        self.time_updates.append([t, car.id, SPEED_CHANGE, self.maxSpeed])
    
    if reply != None:
      # predicted arrival time:
      current = self.length
      key = -1
      for k, v in self.lane.items():
        if v < current:
          key = k
      
      arrival = 0
      for c in self.cars:
        if c.id == key:
          arrival = current/c.v
          break

      if arrival < (reply.time * 1.1):
        # Decelerate (a little bit harder than really needed!)
        new_speed = current/(reply.time * 1.1)
        self.time_updates.append([0, c.id, SPEED_UPDATE, (new_speed + c.v)/2])
        self.time_updates.append([(c.v - new_speed)/c.dv_neg_max, c.id, SPEED_UPDATE, new_speed])
        arrival = reply.time * 1.1
      
      # If we arrive later: Keep this speed, don't accelerate... that's just crazy!
      self.time_updates.append([arrival, c.id, END_OF_ROAD])


  """
  "
  " The internal transisition function.
  "
  """
  def intTransition(self):
    # List of internal transitions:
    #   - Car reached breaking point. (slow down to 30km/h)
    #   - Car reached checkInt point. (check the intersection)
    #   - Car has it's speed changed. (check all cars behind him)
    #   - Car reached end of the road. (push it out)
    update = self.time_updates.pop(0)
    car = None
    for c in self.cars:
      if c.id in self.lane:
        self.lane[c.id] -= c.v * self.elapsed
      
      if c.id == update[1]:
        car = c
    
    for ta in self.time_updates:
      ta[0] -= self.elapsed
    
    if update[2] == BREAKING_POINT:
      # Start breaking: take average of current speed and 30km/h
      self.time_updates.append([0, car.id, SPEED_CHANGE, (car.v + 8.3)/2])
      self.time_updates.append([(car.v - 8.3)/car.dv_neg_max, car.id, CHECKING_POINT])
      car.v = (car.v + 8.3)/2
    elif update[2] == CHECKING_POINT:
      if update[1] in self.lane:
        self.poke(self.query, Query(self.intA))
      self.time_updates.append([0, car.id, SPEED_CHANGE, 8.3])
    elif update[2] == SPEED_CHANGE:
      # Determine new time for leaving.
      car.v = update[3]
      # Determine collision point for the car behind him.
      # Determine the point at which the car has to start breaking.
      if update[1] in self.lane:
        d = self.lane[car.id]
        key = -1
        current = self.length + 1 
        for k, v in self.lane.items():
          if v < current and v > d:
            current = v
            key = k
        
        if key == -1:
          return;
        
        for c in self.cars:
          if k == c.id:
            # Calculate differences
            d_diff = current - d
            v_diff = c.v - car.v
            
            if v_diff < 0:
              # Accelerate to the same speed!
              c.v = (c.v + car.v)/2
              self.time_updates.push([0, c.id, SPEED_CHANGE, (c.v + car.v)/2])
              self.time_updates.push([(car.v - c.v)/c.dv_pos_max, c.id, SPEED_CHANGE, car.v])
            elif v_diff > 0:
              # Deccelerate?
              # The time we can drive at this speed before having to brake. (cp: collision point, bp: breaking point)
              bp = d_diff/v_diff - v_diff/c.dv_neg_max
              self.time_updates.push([bp, c.id, SPEED_CHANGE, (c.v + car.v)/2])
              self.time_updates.push([bp + v_diff/c.dv_neg_max, c.id, SPEED_CHANGE, car.v])
          break
    elif update[2] == END_OF_ROAD:
      if update[1] in self.lane:
        # Push the car into intB
        self.poke(self.car_out, car)
        self.cars.remove(car)


  """
  "
  " The time advance function.
  "
  """
  def timeAdvance(self):
    self.time_updates.sort()
    return self.time_updates[0][0]


"""
"
" The query a road will send to an intersection to determine whether the car
" has to slow down, or even stop.
"
"""
class Query():
  
  """
  "
  " Initializer
  "
  " @param sid: The id of the sender (the ID of the source intersection will be used).
  "
  """
  def __init__(self, sid):
    self.id = sid


"""
"
" The reply an intersection will send as a response to a query.
"
"""
class Reply():
  
  """
  "
  " Initializer
  "
  " @param t: The amount of time the car has to wait before entering the intersection.
  "
  """
  def __init__(self, t):
    self.time = t


"""
"
" This class encapsulates the Intersection.
"
"""
class Intersection(AtomicDEVS):
  
  """
  "
  " Initiator
  "
  " @param ints_in: A list of intersections that lead to this intersection (IDs).
  " @param ints_out: A list of intersections that this intersection leads to (IDs).
  "
  """
  unique_ID = 1
  
  def __init__(self, ints_in, ints_out, name = "None"):
    AtomicDEVS.__init__(self, name = name)
    
    self.id = Intersection.unique_ID
    Intersection.unique_ID += 1
    self.ints_in = ints_in
    self.ints_out = ints_out
    
    self.car_in = self.addInPort(name="Cars_in")
    self.query_in = self.addInPort(name="Queries")
    
    self.car_out = []
    self.reply_out = []
    
    for i in range(len(self.ints_in)):
      self.reply_out.append(self.addOutPort(name="Reply_" + str(self.ints_in[i])))
    
    for i in range(len(self.ints_out)):
      self.car_out.append(self.addOutPort(name="Car_out_" + str(self.ints_in[i])))
    
    self.car = None
    self.occupied = False
    self.reserved = False
    self.queries = []
    
    self.free = 0

  """
  "
  " Connect this intersection with another intersection.
  "
  " @param intersection: The ID of the intersection.
  " @param in_out: string indicating whether it is an "in" or "out" intersection.
  "
  """  
  def addIntersection(self, intersection, in_out):
    if in_out == "in":
      self.ints_in.append(intersection)
      self.reply_out.append(self.addOutPort(name="Reply_" + str(intersection)))
    if in_out == "out":
      self.ints_out.append(intersection)
      self.car_out.append(self.addOutPort(name="Car_out_" + str(intersection)))
  
  """
  "
  " The external transition function
  "
  """
  def extTransition(self):

    self.free -= self.elapsed
    c = self.peek(self.car_in)
    if c != None:
      if self.occupied:
        self.car.v = 0
        c.v = 0
        self.free = INFINITY
      else:
        self.reserved = False
        self.occupied = True
        self.car = c
        # Determine time it takes to leave:
        self.free = 7/c.v
    
    q = self.peek(self.query_in)
    if q != None:
      if self.reserved:
        # Cars must wait indefinatly. (locally store the query)
        self.queries.append([0, Reply(INFINITY), q.id])
      elif self.occupied:
        self.queries.append([0, Reply(self.free), q.id])
        self.reserved = True
      else:
        self.queries.append([0, Reply(0), q.id])
        self.reserved = True
  
  
  """
  "
  " The internal transisition function.
  "
  """
  def intTransition(self):
    # Possible internal transitions:
    #  - Car leaves
    #  - Send message
    
    # Take the one with the smalles value (close to 0)
    chosen_message = None
    t_message = INFINITY
    for message in self.queries:
      message[0] -= self.elapsed
      if message[0] < t_message:
        chosen_message = message
    
    if self.occupied:
      self.free -= self.elapsed
    
    if self.occupied and self.free < t_message:
      # Make car leave
      intersection = self.car.trajectory.pop(0)
      self.poke(self.car_out[self.ints_out.index(intersection)], self.car)
      self.car = None
      self.occupied = False
    elif chosen_message != None:
      # Handle message
      self.queries.remove(chosen_message)
      print chosen_message[2]
      print self.id
      self.poke(self.reply_out[self.ints_in.index(chosen_message[2])], chosen_message[1])
  
  """
  "
  " The time advance function.
  "
  """
  def timeAdvance(self):
    if self.occupied:
      t_free = self.free
    else:
      t_free = INFINITY
    
    t_messages = INFINITY
    for message in self.queries:
      if message[0] < t_messages:
        t_messages = message[0]
      
    return min(t_free, t_messages)

