"""
"
" @File: main.py
" @Description: The main file. This is the starting point of the entire map generation.
" @Author: Chris Vesters
" @Date: 26/12/2012
"
"""
import networkx
import igraph

from t_core.messages import Packet, MatchSet
from t_core.matcher import Matcher
from t_core.iterator import Iterator
from t_core.rewriter import Rewriter

from HInitial import HInitial
from HExpandNorthLHS import HExpandNorthLHS
from HExpandNorthRHS import HExpandNorthRHS
from HExpandEastLHS import HExpandEastLHS
from HExpandEastRHS import HExpandEastRHS
from HExpandSouthLHS import HExpandSouthLHS
from HExpandSouthRHS import HExpandSouthRHS
from HExpandWestLHS import HExpandWestLHS
from HExpandWestRHS import HExpandWestRHS
from HConnectIntersectionsLHS import HConnectIntersectionsLHS
from HConnectIntersectionsRHS import HConnectIntersectionsRHS
from HOneWayLHS import HOneWayLHS
from HOneWayRHS import HOneWayRHS

from HGrowRoadLHS import HGrowRoadLHS
from HGrowRoadRHS import HGrowRoadRHS
from HOneWayLHS import HOneWayLHS
from HOneWayRHS import HOneWayRHS


"""
"
" This method will simplify the graph by removing unneeded nodes.
"
" @param graph: The graph we will simplify
"
" @return The simplified version of the graph.
"
"""
def simplify(graph):
  nodes = graph.vs.select(mm__ = "Intersection")
  nodesIndex = []
  for node in nodes:
    nodesIndex.append(node.index)
  edges = []
  weights = []
  
  for n in nodes:
    indices = [n.index]
    visited = [n.index]
    
    weight = 0
    while len(indices) > 0:
      for edge in graph.es.select(_source = indices.pop()):
		# We do not have to check whether the target has already been visited.
		# It is enough to check whether the index is the same as the current one.
		# This is because we can only get to the next intersection and not any further!
        if ((edge.target in nodesIndex) and (edge.target != n.index)):
          edges.append((n.index, edge.target))
          weights.append(weight)
          weight = 0
        elif edge.target not in visited:
          # Check for the special case in which it is a road:
          if graph.vs[edge.target]["mm__"] == "Road":
            weight += graph.vs[edge.target]["length"]
          indices.append(edge.target)
        visited.append(edge.target)
  
  # Add nodes.
  g = igraph.Graph(len(nodes))
  g.vs["index"] = range(len(nodes))
  g.vs["x"] = nodes["x"]
  g.vs["y"] = nodes["y"]
  
  # Make the graph directed!
  g.to_directed()
  
  # Add edges.
  # Note that we need to fix the id's!
  for i, node in enumerate(nodes):
    for j, edge in enumerate(edges):
      if edge[0] == node.index:
        edges[j] = (i, edge[1])
      
      if edge[1] == node.index:
        edges[j] = (edge[0], i)
  
  g.add_edges(edges)
  g.es["weight"] = weights
  
  return g


"""
"
" This method will convert the graph from iGraph to networkX
"
" @param graph: The graph we will convert
"
" @return The graph in the form of a networkx Graph!
"
"""
def toNetworkX(graph):
  g = networkx.DiGraph()
  
  for node in graph.vs:
    g.add_node(node.index, x = node["x"], y = node["y"])
  
  for edge in graph.es:
    g.add_edge(edge.source, edge.target, edge.attributes())
  
  return g


"""
"
" This method will check the feasibility of the graph.
"
" @param graph: The graph we need to check (NetworkX format).
"
" @return True if the graph is feasible, False if not.
"
"""
def feasible(graph):
  # There may not be any intersection from which we can't reach every other intersection!
  return networkx.is_strongly_connected(graph)


"""
"
" This method evaluates a graph.
"
" @param graph: The graph (networkX) we will evaluate.
"
" @return A numerical value indicating how good the graph is.
"
"""
def evaluate(graph, size):
  # Edit the graph: the weight of each edge should be increased with 7 (distance of the intersection) and the amount of incomming
  # edges/roads the intersection has. This is because we want to minimize the amount of intersections on our way.
  if len(graph) == 1:
    return 0
  
  ec = networkx.eccentricity(graph)
  diameter = networkx.diameter(graph, ec)
  
  for source in graph.nodes_iter():
    for in_edge in graph.in_edges_iter(source, True):
      in_edge[2]["capacity"] = 1
    for out_edge in graph.out_edges_iter(source, True):
      out_edge[2]["capacity"] = 1
  
  """ This code isn't used!
  # Penalty for intersections!
  for source in graph.nodes_iter():
    d = 0.0
    if len(graph.in_edges(source)) > 2:
      d += 14
    else:
      d += 10.5
    d /= 2

    for in_edge in graph.in_edges_iter(source, True):
      in_edge[2]["weight"] += d
      in_edge[2]["capacity"] = 1
    for out_edge in graph.out_edges_iter(source, True):
      out_edge[2]["weight"] += d
      out_edge[2]["capacity"] = 1
  
  # Calculate the average of the shortest path for each node to the furthest node.
  # Take the shortest path for each node to every other node. Search for the biggest shortest path.
  path = networkx.shortest_path_length(graph, weight = 'weight')
  total_rel_dist = 0
  for vertex in path:
    source = graph.node[vertex]
    longest_rel_dist = 0
    for key, value in path[vertex].iteritems():
      if (key != vertex):
        target = graph.node[key]
        man_dist = abs(source['x'] - target['x']) + abs(source['y'] - target['y'])
        longest_rel_dist = max(longest_rel_dist, value/man_dist)
    
    total_rel_dist += longest_rel_dist
  """
  
  # How many roads are there to get to any intersection?
  total_min_cut = 0
  for s in graph.nodes():
    for t in graph.nodes():
      if (s != t):
        total_min_cut += networkx.min_cut(graph, s, t)
  
  # We have added n*(n-1) min_cuts, divide by this number to get the average!
  min_cut = total_min_cut / (len(graph) * (len(graph) - 1))
  diam_value = diameter - 2 * size
  return min_cut/((diam_value * diam_value) + 1)

  
"""
"
" This method will draw the graph.
"
" @param graph: The graph we need to draw.
" @param plot: A boolean indicating if we want to plot the graph.
" @param size: The size of the image.
" @param name: The name of the file we will save the image too.
"
"""
def draw(graph, size, scale = 1, plot=False, name="graph.svg"):
  visual_style = {}
  visual_style["layout"] = 'fr'
  visual_style["margin"] = 10
  visual_style["layout"] = 'auto'
  
  shapes = []
  for vertex in graph.vs:
    vertex['x'] *= scale
    vertex['y'] *= scale
    shapes.append(1)
  
  # We add 4 extra vertices in the corners. These vertices are hidden!
  # The only reason for this is to prevent iGraph from rescaling.
  graph.add_vertices(4)
  
  count = graph.vcount()
  
  graph.vs[count - 4]["x"] = 0
  graph.vs[count - 4]["y"] = 0
  shapes.append(0)
  
  graph.vs[count - 3]["x"] = size * scale
  graph.vs[count - 3]["y"] = 0
  shapes.append(0)
  
  graph.vs[count - 2]["x"] = 0
  graph.vs[count - 2]["y"] = size * scale
  shapes.append(0)
  
  graph.vs[count - 1]["x"] = size * scale
  graph.vs[count - 1]["y"] = size * scale
  shapes.append(0)
  
  if plot:
    igraph.plot(graph, bbox=(0, 0, size * scale, size * scale), **visual_style)
  
  graph.write_svg("imgs/" + name, 'auto', size * scale, size * scale, labels = None, shapes = shapes, vertex_size = 3.5)


"""
"
" This class will always execute the best rules.
"
"""
class HillClimb():

  """
  "
  " Initialize the BestRule.
  "
  " @param LHS: a set of LHS.
  " @param RHS: a set of RHS.
  " # @param n: the number of best rules we will examine.
  "
  """
  def __init__(self, LHS, RHS, initial_packet):
    assert len(LHS) == len(RHS), "LHS and RHS are not of the same size!"
    
    self.M = []
    self.W = []
    for i in range(len(LHS)):
      self.M.append(Matcher(condition = LHS[i]))
      self.W.append(Rewriter(condition = RHS[i]))
      
    self.I = Iterator()
    
    self.size = initial_packet.graph.vs.find(mm__ = "MapInformation")["size"]
    nx = toNetworkX(simplify(initial_packet.graph))
    self.candidates = [(evaluate(nx, self.size), initial_packet)]
    self.is_success = True
    
    draw(simplify(initial_packet.graph), self.size, scale = 3, name = "initial.svg")


  """
  "
  " Run the HillClimb algorithm.
  "
  """
  def run(self):
    i = 0
    best = (0, None)
    while len(self.candidates) > 0:
      self.candidates.sort()
      current = self.candidates.pop()
      score = current[0]
      packet = current[1]
      packet_original = packet.clone()
      print "CURRENT CANDIDATES: " + str(len(self.candidates))
      
      new = False
      # Generate all children.
      for matcher in self.M:
        packet = matcher.packet_in(packet)
        if not matcher.is_success:
          continue
        
        rewriter = self.W[self.M.index(matcher)]
        packet = self.I.packet_in(packet)
        while self.I.is_success:
          packet_copied = packet.clone()
          packet = rewriter.packet_in(packet)
          packet.match_sets = {}
          simplified = simplify(packet.graph)
          nx = toNetworkX(simplified)
          
          # Feasibility check:
          if not feasible(nx):
            packet = self.I.next_in(packet_copied)
            continue
          
          # Evaluation:
          new_score = evaluate(nx, self.size)
          
          if new_score >= score:
            new = True
            self.candidates.append((new_score, packet))
          
          packet = self.I.next_in(packet_copied)
      
      if not new and score > best[0]:
        best = (score, packet_original)
        # The graph is saved as an intermediate result!
        i += 1
        simplified = simplify(packet_original.graph)
        export(simplified, name = "local_best-" + str(i) + ".svg")
        draw(simplified, self.size, scale = 3, name = "local_best-" + str(i) + ".svg")
    
    # This is the best match!
    simplified = simplify(best[1].graph)
    nx = toNetworkX(simplified)
    export(simplified, name = "best-" + str(best[0])) + ".svg"
    draw(simplified, self.size, scale = 3, name = "best-" + str(best[0]) + ".svg")


"""
"
" This method will create a PyDevs experiment file.
"
" @param graph: The graph.
"
"""
def export(graph, name = "RoadNetworkExperiment.py"):
  spaces = ""
  # We only generate the experiment, the model file is provided!
  f = open(name, 'w')
  f.write("# =================================================== #\n")
  f.write("#                                                     #\n")
  f.write("# This file has been automaticaly generated!          #\n")
  f.write("# @file: RoadNetworkExperiment.py                     #\n")
  f.write("# @author: Chris Vesters                              #\n")
  import datetime
  now = datetime.datetime.now()
  f.write("# @date: " + now.strftime("%d/%m/%Y") + "                                   #\n" )
  f.write("#                                                     #\n")
  f.write("# Note: it may be helpful to set the path to pydevs.  #\n")
  f.write("# This can be done by:                                #\n")
  f.write("# import sys                                          #\n")
  f.write("# import os.path                                      #\n")
  f.write("# sys.path.append(os.path...)                         #\n")
  f.write("#                                                     #\n")
  f.write("# =================================================== #\n\n\n")
  f.write("# Import code for model simulation:\n")
  f.write("import pydevs\n")
  f.write("from pydevs.infinity import *\n")
  f.write("from pydevs.simulator import *\n")
  f.write("from RoadNetworkModel import *\n\n\n")
  f.write("\"\"\"\n\"\n\" This class represents the entire road network.\n\"\n\"\"\"\n")
  f.write("class Simulation(CoupledDEVS):\n\n")
  spaces += "  "
  f.write(spaces + "\"\"\"\n" + spaces + "\"\n" + spaces +"\" Initializer.\n" + spaces + "\"\n" + spaces + "\" @param name: The name of the DEVS module.\n" + spaces + "\"\n" + spaces + "\"\"\"\n")
  f.write(spaces + "def __init__(self, name = None):\n")
  spaces += "  "
  f.write(spaces + "CoupledDEVS.__init__(self, name)\n\n")
  f.write(spaces + "intersections = []\n")
  
  for vertex in graph.vs:
    f.write(spaces + "intersections.append(self.addSubModel(Intersection([], [], name = \"Intersection\")))\n")
  f.write(spaces + "\n")
  
  f.write(spaces + "roads = []\n")
  for edge in graph.es:
    f.write(spaces + "roads.append(self.addSubModel(Road(intersections[" + str(edge.source) + "].id, intersections[" + str(edge.target) + "].id, " + str(edge["weight"]) + ", 13.8, name = \"Road\")))\n")
    f.write(spaces + "intersections[" + str(edge.source) + "].addIntersection(intersections[" + str(edge.target) + "].id, \"out\")\n")
    f.write(spaces + "intersections[" + str(edge.target) + "].addIntersection(intersections[" + str(edge.source) + "].id, \"in\")\n")
    f.write(spaces + "i = intersections[" + str(edge.source) + "].ints_out.index(intersections[" + str(edge.target) + "].id)\n")
    f.write(spaces + "self.connectPorts(intersections[" + str(edge.source) + "].car_out[i], roads[-1].car_in)\n")
    f.write(spaces + "self.connectPorts(roads[-1].query, intersections[" + str(edge.target) + "].query_in)\n")
    f.write(spaces + "i = intersections[" + str(edge.target) + "].ints_in.index(intersections[" + str(edge.source) + "].id)\n")
    f.write(spaces + "self.connectPorts(intersections[" + str(edge.target) + "].reply_out[i], roads[-1].reply)\n")
    f.write(spaces + "self.connectPorts(roads[-1].car_out, intersections[" + str(edge.target) + "].car_in)\n\n")
  
  f.write("\n")
  spaces = spaces[2:]  

  f.write(spaces + "\"\"\"\n" + spaces + "\"\n" + spaces + "\" The select function.\n" + spaces + "\"\n" + spaces + "\"\"\"\n")
  f.write(spaces + "def select(self, immList):\n")
  spaces += "  "
  f.write(spaces + "return immList[-1]\n\n")
  spaces = spaces[4:]

  f.write(spaces + "system = Simulation(name = \"System\")\n")
  f.write(spaces + "sim = Simulator(system)\n\n")
  f.write(spaces + "\"\"\"\n" + spaces + "\"\n" + spaces + "\" This method specifies the terminating condition for the simulation.\n" + spaces + "\"\n" + spaces + "\"\"\"\n")
  f.write(spaces + "def terminate_whenEndTimeReached(model, clock, end_time = 100):\n")
  spaces += "  "
  f.write(spaces + "if clock >= end_time:\n")
  spaces += "  "
  f.write(spaces + "return True\n")
  spaces = spaces[2:]
  f.write(spaces + "else:\n")
  spaces += "  "
  f.write(spaces + "return False\n\n")
  spaces = spaces[4:]
  f.write(spaces + "sim.simulate(termination_condition = terminate_whenEndTimeReached, verbose = True)\n")


# NOTE: we actually want all possible matches, not just one for each rule!
# A matcher for each rule.
packet = Packet()
packet.graph = HInitial()

hill = HillClimb(
  [HExpandNorthLHS(), HExpandEastLHS(), HExpandSouthLHS(), HExpandWestLHS(), HGrowRoadLHS(), HConnectIntersectionsLHS(), HOneWayLHS()],
  [HExpandNorthRHS(), HExpandEastRHS(), HExpandSouthRHS(), HExpandWestRHS(), HGrowRoadRHS(), HConnectIntersectionsRHS(), HOneWayRHS()],
  packet)

hill.run()
